from multiprocessing import get_context

ctx = get_context('fork')


class Process(ctx.Process):
    def __init__(self, *args, **kwargs):
        super(Process, self).__init__(*args, **kwargs)

        self._pconn, self._cconn = ctx.Pipe(duplex=False)
        self._exception = None

    def run(self):
        try:
            super(Process, self).run()

            self._cconn.send(None)
        except Exception as e:
            self._cconn.send(e)
        finally:
            self._cconn.close()
            self._pconn.close()

    @property
    def exception(self):
        if self._pconn.poll():
            self._exception = self._pconn.recv()

        return self._exception


__all__ = [
    'Process',
    'ctx'
]



import json
import re
import traceback

from collections import OrderedDict

from datetime import datetime, timezone

import inflect
import numpy
import pytz

from ..api.appd import AppD
from ..api.extrahop import ExtraHop
from ..api.newrelic import NewRelic
from ..database import DatabaseAccessLayer

from ...main import *
from ..util.multiprocessing_helper import *


def aggregate_values_by_func(func, values):
    if is_number(func):
        value = numpy.percentile(values, int(func))
    else:
        value = 0

        if func == 'avg':
            if values:
                value = numpy.average(values)
        elif func == 'count':
            value = len(values)
        elif func == 'min':
            value = numpy.min(values)
        elif func == 'max':
            value = numpy.max(values)
        elif func == 'sum':
            value = numpy.sum(values)

    return value


def format_label(string):
    label = re.sub(r'[().\'\"]|((?<![/:])/)', '', string)
    label = re.sub(r'^%(?=\w)', 'percent ', label)
    label = re.sub(r'%', ' percent', label)
    label = re.sub(r'>', 'greater than', label)
    label = re.sub(r'<', 'less than', label)
    label = re.sub(r'SQL\*Net', 'Net8', label)
    label = re.sub(r'[ \-]', '_', label)
    label = re.sub(r'\W', '', label)
    label = re.sub(r'_{2,}', '_', label)

    return label.lower()


def format_metric(string):
    metric = re.sub(r'(?<=\S)(/sec)', ' per sec', string)
    metric = re.sub(r'[ \-]', '_', metric)
    metric = re.sub(r'\W', '', metric)
    metric = re.sub(r'_{2,}', '_', metric)

    return metric.lower()


def get_metrics(func, job_name, args, push_services, conn=None):
    collector_metrics = func(*args)

    if conn is not None:
        conn.send(collector_metrics)

    aps.app.logger.info("Fetched metrics for job '" + job_name + "'.")

    set_job_collector_metrics(job_name, collector_metrics, push_services)

    for vendor, services in push_services.items():
        globals()['push_' + vendor + '_job_metrics'](job_name, services)


def get_metric_info(service):
    service_name_pattern = re.compile(r'\[(?P<instance>[^\[\]]+)\]', re.X)

    prefix = re.sub(service_name_pattern, '', service)

    instance = None

    m = service_name_pattern.search(service)

    if m is not None:
        components = m.groupdict()

        instance = components['instance']

    return prefix.lower(), instance


def is_element_in_iterable_no_case(element, iterable):
    return element.lower() in [val.lower() for val in list(iterable)]


def is_number(n):
    try:
        float(n)
    except ValueError:
        return False

    return True


def make_label_singular(label):
    p = inflect.engine()

    words = label.split('_')

    words.append(p.singular_noun(words.pop()))

    return '_'.join(words)


def test_aggregation_match(value, aggregation):
    if 'threshold' in aggregation.keys():
        return eval('%d %s %d' % (value, aggregation['threshold']['operator'], int(aggregation['threshold']['value'])))

    return True


def test_aggregation_settings(aggregation, job_name):
    if 'funcs' in aggregation.keys():
        valid_funcs = [
            'avg',
            'count',
            'min',
            'max',
            'sum'
        ]

        for func in aggregation['funcs']:
            if func not in valid_funcs:
                if is_number(func):
                    if int(func) not in range(0, 101):
                        raise ValueError("Invalid aggregation percentile '" + func + "' in job '" + job_name + "'.")
                else:
                    raise ValueError("Unsupported aggregation function '" + func + "' in job '" + job_name + "'.")
    else:
        aggregation['funcs'] = [
            'count'
        ]

    if 'threshold' in aggregation.keys():
        if 'operator' in aggregation['threshold'].keys():
            valid_operator = [
                '>',
                '<',
                '>=',
                '<=',
                '=',
                '<>',
                '!='
            ]

            if aggregation['threshold']['operator'] not in valid_operator:
                raise ValueError("Unsupported aggregation threshold operator '" + aggregation['threshold']['operator'] + "' in job '" + job_name + "'.")
        else:
            raise ValueError("No operator specified for aggregation threshold in job '" + job_name + "'.")

        if 'value' in aggregation['threshold'].keys():
            if is_number(aggregation['threshold']['value']):
                aggregation['threshold']['value'] = aggregation['threshold']['value']
            else:
                raise ValueError("Invalid value '" + aggregation['threshold']['value'] + "' specified for aggregation threshold in job '" + job_name + "'.")
        else:
            raise ValueError("No value specified for aggregation threshold in job '" + job_name + "'.")

    return aggregation


def to_string(n):
    return str(n) if n is not None else ''


class Metrics:
    @staticmethod
    def _get_appdynamics_metrics(job_name, service_names, application, metric_path, minutes, static_labels=()):
        collector_metrics = {}

        with aps.app.app_context():
            all_labels = []

            metric_data = {}

            for service_name in service_names['source']:
                prefix, instance = get_metric_info(service_name)

                dal = AppD(aa)
                dal.init_aa(service_name)

                aps.app.logger.info("Initialized connection for job '" + job_name + "' to AppDynamics service '" + service_name + "'.")

                if not application and dal.application is not None:
                    application = dal.application

                tiers = []
                nodes = []

                if application != 'Database Monitoring':
                    result = dal.client.get_tiers(application)

                    for tier_obj in result:
                        tiers.append(tier_obj.name)

                    result = dal.client.get_nodes(application)

                    for node_obj in result:
                        nodes.append(node_obj.name)

                options = {
                    'time_range_type': 'BEFORE_NOW',
                    'duration_in_mins': minutes,
                    'rollup': True
                }

                result = dal.client.get_metrics(metric_path, application, **options)

                metrics = []

                for metric_obj in result:
                    if metric_obj.values:
                        components = metric_obj.path.split('|')

                        if components is not None:
                            metric_profile = [
                                format_label(components.pop(0)),
                                format_label(components.pop())
                            ]

                            label_dict = OrderedDict([
                                ('application', application)
                            ])

                            if service_names['push'] or aps.app.config['DEFAULT_PUSH_SERVICE_NAMES']:
                                if instance is not None:
                                    label_dict['instance'] = instance

                            if application == 'Database Monitoring':
                                label = make_label_singular(metric_profile[0])

                                label_dict[label] = components.pop(0)

                                if ':' in components[0]:
                                    label_dict['node'] = components.pop(0)

                                for component in components:
                                    metric_profile.insert(len(metric_profile)-1, format_label(component))
                            else:
                                if metric_profile[0] == 'application_infrastructure_performance':
                                    i = 1

                                    while i < len(components):
                                        if components[i] != 'Individual Nodes' and components[i] not in nodes:
                                            metric_profile.insert(len(metric_profile)-1, format_label(components[i]))

                                            if components[i-1] == 'JVM':
                                                if components[i] == 'Garbage Collection':
                                                    label = make_label_singular(format_label(components.pop(i+1)))

                                                    label_dict[label] = components.pop(i+1)
                                                elif components[i] == 'Memory':
                                                    label = format_label(components[i])

                                                    label_dict[label] = components.pop(i+1)

                                        i += 1
                                else:
                                    if metric_profile[0] == 'business_transaction_performance':
                                        label = make_label_singular(format_label(components.pop(0)))

                                        if label == 'business_transaction_group':
                                            label_dict[label] = components.pop(0)
                                        elif label == 'business_transaction':
                                            label_dict[label] = components.pop(1)
                                    elif metric_profile[0] in ['errors', 'information_points', 'service_endpoints']:
                                        label = make_label_singular(metric_profile[0])

                                        label_dict[label] = components.pop(1)

                                    external_calls_index = 0

                                    while len(components) > 0:
                                        if components[0] in tiers:
                                            label_dict['tier'] = components.pop(0)
                                        else:
                                            pattern = re.compile(r'^(?P<desc>Discovered backend call) - (?P<entity>.+)$')

                                            m = pattern.match(components[0])

                                            if m is not None:
                                                subcomponents = m.groupdict()

                                                label = format_label(subcomponents['desc'])

                                                label_dict[label] = subcomponents['entity']

                                                components.pop(0)
                                            else:
                                                label = make_label_singular(format_label(components.pop(0)))

                                                if label == 'external_call':
                                                    external_calls_index += 1

                                                    label_base = '%s_%d' % (label, external_calls_index)

                                                    pattern = re.compile(r'^Call-(?P<type>\S+) to (?P<target>.+)$')

                                                    m = pattern.match(components.pop(0))

                                                    if m is not None:
                                                        subcomponents = m.groupdict()

                                                        label_dict[label_base] = subcomponents['type']

                                                        pattern = re.compile(r'^(?P<desc>Discovered backend call) - (?P<entity>.+)$')

                                                        m = pattern.match(subcomponents['target'])

                                                        if m is not None:
                                                            subcomponents = m.groupdict()

                                                            label = '%s_%s' % (label_base, format_label(subcomponents['desc']))

                                                            label_dict[label] = subcomponents['entity']
                                                        else:
                                                            label = '%s_tier' % label_base

                                                            label_dict[label] = subcomponents['target']

                                                            if len(components) > 0:
                                                                components.pop(0)
                                                elif label == 'incoming_cross_app_call':
                                                    label_base = label

                                                    label = '%s_application' % label_base

                                                    label_dict[label] = components.pop(0)

                                                    pattern = re.compile(r'^Call-(?P<type>\S+) from (?P<source>.+)$')

                                                    m = pattern.match(components.pop(0))

                                                    if m is not None:
                                                        subcomponents = m.groupdict()

                                                        label_dict[label_base] = subcomponents['type']

                                                        label = '%s_tier' % label_base

                                                        label_dict[label] = subcomponents['source']
                                                elif label in ['individual_node', 'thread_task']:
                                                    label_dict[label] = components.pop(0)

                            row = metric_obj.values[0].__dict__

                            metric_dict = OrderedDict([
                                (key, value) for key, value in row.items() if key != 'start_time_ms'
                            ])

                            if not metrics:
                                metrics = [
                                    metric for metric, value in metric_dict.items() if is_number(value)
                                ]

                            label_dict.update(OrderedDict([
                                (format_label(label), value) for label, value in static_labels if format_label(label) not in label_dict.keys()
                            ]))

                            for label in label_dict.keys():
                                if label not in all_labels:
                                    all_labels.append(label)

                            timestamp = row['start_time_ms'] / 1000 + (int(minutes) * 60)

                            for metric in metrics:
                                metric_name = '%s.%s.%s' % (prefix, '.'.join(metric_profile), metric)

                                if metric_name not in metric_data.keys():
                                    metric_data[metric_name] = []

                                metric_data[metric_name].append((
                                    label_dict,
                                    int(metric_dict[metric]),
                                    timestamp
                                ))

            for metric_name, data in metric_data.items():
                for (label_dict, value, timestamp) in data:
                    normalized_label_dict = OrderedDict()

                    for label in all_labels:
                        if label in label_dict.keys():
                            normalized_label_dict[label] = label_dict[label]
                        else:
                            normalized_label_dict[label] = ''

                    json_label_data = json.dumps(normalized_label_dict)

                    if metric_name not in collector_metrics.keys():
                        collector_metrics[metric_name] = {}

                    if json_label_data not in collector_metrics[metric_name].keys():
                        collector_metrics[metric_name][json_label_data] = (value, timestamp)

        return collector_metrics

    @staticmethod
    def _get_database_metrics(job_name, service_names, statement, value_columns, static_labels=(), timestamp_column=None, timezones={}):
        def is_value_column(column):
            return is_element_in_iterable_no_case(column, value_columns)

        collector_metrics = {}

        with aps.app.app_context():
            for service_name in service_names['source']:
                prefix, instance = get_metric_info(service_name)

                dal = DatabaseAccessLayer(db)
                dal.init_db(service_name)

                aps.app.logger.info("Initialized connection for job '" + job_name + "' to database service '" + service_name + "'.")

                result = dal.execute(statement)

                timestamp = datetime.now(timezone.utc).timestamp()

                db_tzinfo = None

                if timestamp_column is not None and service_name in timezones.keys():
                    db_tzinfo = pytz.timezone(timezones[service_name])

                normalized_value_columns = None
                normalized_timestamp_column = None

                for row in result:
                    def is_unmatched_value_column(column):
                        return not is_element_in_iterable_no_case(column, row.keys())

                    if normalized_value_columns is None:
                        normalized_value_columns = list(filter(is_value_column, row.keys()))

                        unmatched_value_columns = list(filter(is_unmatched_value_column, value_columns))

                        if unmatched_value_columns:
                            raise ValueError("Value column(s) " + ", ".join(["'" + column + "'" for column in unmatched_value_columns]) + " specified in job '" + job_name + "' not returned in query result.")

                    if timestamp_column is not None and normalized_timestamp_column is None:
                        for column in row.keys():
                            if timestamp_column.lower() == column.lower():
                                normalized_timestamp_column = column

                        if normalized_timestamp_column is None:
                            raise ValueError("Timestamp column '" + timestamp_column + "' specified in job '" + job_name + "' not returned in query result.")

                    label_dict = OrderedDict()

                    if service_names['push'] or aps.app.config['DEFAULT_PUSH_SERVICE_NAMES']:
                        if instance is not None:
                            label_dict['instance'] = instance

                    label_dict.update(OrderedDict([
                        (format_label(column), to_string(row[column])) for column in row.keys() if column not in normalized_value_columns and column != normalized_timestamp_column
                    ]))

                    label_dict.update(OrderedDict([
                        (format_label(label), value) for label, value in static_labels if format_label(label) not in label_dict.keys()
                    ]))

                    if db_tzinfo is not None:
                        if isinstance(row[normalized_timestamp_column], datetime):
                            if row[normalized_timestamp_column].tzinfo is not None and row[normalized_timestamp_column].tzinfo.utcoffset(row[normalized_timestamp_column]) is not None:
                                timestamp = row[normalized_timestamp_column].timestamp()
                            else:
                                timestamp = row[normalized_timestamp_column].astimezone(db_tzinfo).timestamp()

                    json_label_data = json.dumps(label_dict)

                    for column in normalized_value_columns:
                        metric_name = '%s.%s' % (prefix, format_metric(column))

                        if metric_name not in collector_metrics.keys():
                            collector_metrics[metric_name] = {}

                        if json_label_data not in collector_metrics[metric_name].keys():
                            collector_metrics[metric_name][json_label_data] = (row[column], timestamp)

        return collector_metrics

    @staticmethod
    def _get_extrahop_metrics(job_name, service_names, params, metric, aggregation, minutes, static_labels=()):
        def _get_metrics():
            _metrics = []

            options = {**params, **{
                'from': '-%dm' % minutes,
                'until': 0
            }}

            _result = dal.client.get_metrics(**options)

            if 'xid' in _result.keys():
                xid = _result['xid']
                num_results = _result['num_results']

                for i in range(num_results):
                    _result = dal.client.get_metrics_by_xid(xid)

                    if 'stats' in _result.keys():
                        _metrics.append(_result['stats'])
            elif 'stats' in _result.keys():
                _metrics.append(_result['stats'])

            return _metrics

        collector_metrics = {}

        aggregation = test_aggregation_settings(aggregation, job_name)

        with aps.app.app_context():
            for service_name in service_names['source']:
                prefix, instance = get_metric_info(service_name)

                dal = ExtraHop(aa)
                dal.init_aa(service_name)

                aps.app.logger.info("Initialized connection for job '" + job_name + "' to ExtraHop service '" + service_name + "'.")

                result = _get_metrics()

                timestamp = datetime.now(timezone.utc).timestamp()

                if result:
                    metric_dict = OrderedDict()

                    for stats in result:
                        for i in stats:
                            for value in i['values']:
                                if isinstance(value, list):
                                    for j in value:
                                        if 'key' in j.keys():
                                            if j['key']['key_type'] == 'string':
                                                metric_spec_name = j['key']['str']
                                                value = int(j['value'])

                                                if metric_spec_name not in metric_dict.keys():
                                                    metric_dict[metric_spec_name] = []

                                                if test_aggregation_match(value, aggregation):
                                                    metric_dict[metric_spec_name].append(value)

                    for metric_spec_name, values in metric_dict.items():
                        label_dict = OrderedDict([
                            ('metric_spec_name', metric_spec_name)
                        ])

                        if service_names['push'] or aps.app.config['DEFAULT_PUSH_SERVICE_NAMES']:
                            if instance is not None:
                                label_dict['instance'] = instance

                        label_dict.update(OrderedDict([
                            (format_label(label), value) for label, value in static_labels if format_label(label) not in label_dict.keys()
                        ]))

                        json_label_data = json.dumps(label_dict)

                        for func in aggregation['funcs']:
                            metric_name = '%s.%s.%s' % (prefix, format_metric(metric), func.lower())

                            if metric_name not in collector_metrics.keys():
                                collector_metrics[metric_name] = {}

                            if json_label_data not in collector_metrics[metric_name].keys():
                                collector_metrics[metric_name][json_label_data] = (aggregate_values_by_func(func, values), timestamp)

        return collector_metrics

    @staticmethod
    def _get_newrelic_metrics(job_name, service_names, account_id, statement, value_attrs, minutes, static_labels=()):
        def is_value_attr(attr):
            return is_element_in_iterable_no_case(attr, value_attrs)

        collector_metrics = {}

        with aps.app.app_context():
            nrql_query = re.sub(r'[\r\n]+', ' ', statement)

            for service_name in service_names['source']:
                prefix, instance = get_metric_info(service_name)

                dal = NewRelic(aa)
                dal.init_aa(service_name)

                aps.app.logger.info("Initialized connection for job '" + job_name + "' to New Relic service '" + service_name + "'.")

                account_id = account_id or dal.account_id

                query_str = """{
                   actor {
                      account(id: %d) {
                         nrql(query: "%s SINCE %d minutes AGO") {
                            results
                         }
                      }
                   }
                }""" % (account_id, nrql_query, minutes)

                result = dal.client.query(query_str)

                timestamp = datetime.now(timezone.utc).timestamp()

                if 'data' in result.keys():
                    if result['data']['actor']['account']['nrql']['results']:
                        normalized_value_attrs = None

                        for row in result['data']['actor']['account']['nrql']['results']:
                            def is_unmatched_value_attr(attr):
                                return not is_element_in_iterable_no_case(attr, row.keys())

                            if normalized_value_attrs is None:
                                normalized_value_attrs = list(filter(is_value_attr, row.keys()))

                                unmatched_value_attrs = list(filter(is_unmatched_value_attr, value_attrs))

                                if unmatched_value_attrs:
                                    raise ValueError("Value attribute(s) " + ", ".join(["'" + attr + "'" for attr in unmatched_value_attrs]) + " specified in job '" + job_name + "' not returned in query result.")

                            label_dict = OrderedDict()

                            if service_names['push'] or aps.app.config['DEFAULT_PUSH_SERVICE_NAMES']:
                                if instance is not None:
                                    label_dict['instance'] = instance

                            label_dict.update(OrderedDict([
                                (format_label(attr), to_string(row[attr])) for attr in row.keys() if attr not in normalized_value_attrs
                            ]))

                            label_dict.update(OrderedDict([
                                (format_label(label), value) for label, value in static_labels if format_label(label) not in label_dict.keys()
                            ]))

                            json_label_data = json.dumps(label_dict)

                            for attr in normalized_value_attrs:
                                metric_name = '%s.%s' % (prefix, format_metric(attr))

                                if metric_name not in collector_metrics.keys():
                                    collector_metrics[metric_name] = {}

                                if json_label_data not in collector_metrics[metric_name].keys():
                                    collector_metrics[metric_name][json_label_data] = (row[attr], timestamp)

        return collector_metrics

    @staticmethod
    def generate_metrics(*args):
        category = args[0]
        job_name = args[1]
        service_names = args[2]

        job = aps.get_job(job_name)

        func_name = '_get_' + category + '_metrics'

        if hasattr(Metrics, func_name):
            func = getattr(Metrics, func_name)

            with aps.app.app_context():
                try:
                    with JOB_EXECUTION_TIME.labels(job_name, category).time():
                        push_services = get_push_services(
                            job_name,
                            (aps.app.config['DEFAULT_PUSH_SERVICE_NAMES'] + service_names['push'])
                        )

                        enforce_timeout = aps.app.config['LIMIT_JOB_EXECUTION_TIME'] and job.next_run_time is not None

                        if enforce_timeout:
                            # Prevent database jobs with JDBC connections from running under multiprocessing
                            if category == 'database':
                                for service_name in service_names['source']:
                                    if service_name in aps.app.config['SQLALCHEMY_BINDS'].keys():
                                        bind = aps.app.config['SQLALCHEMY_BINDS'][service_name]

                                        result = re.search(r'jdbc', bind)

                                        enforce_timeout = result is None
                                        break

                        if enforce_timeout:
                            pconn, cconn = ctx.Pipe()

                            p = Process(
                                target=get_metrics,
                                args=(func, job_name, args[1:], push_services, cconn)
                            )

                            end_dt = job.next_run_time
                            job_tz = pytz.timezone(end_dt.tzname())

                            p.start()

                            max_execution_time = (end_dt - datetime.now(job_tz)).total_seconds() - 1

                            p.join(timeout=max_execution_time)

                            if pconn.poll():
                                collector_metrics = pconn.recv()

                                set_job_collector_metrics(job_name, collector_metrics, push_services)

                            pconn.close()
                            cconn.close()

                            if p.is_alive():
                                p.kill()

                                raise TimeoutWarning("Process timed out.")

                            if p.exception is not None:
                                raise p.exception
                        else:
                            get_metrics(func, job_name, args[1:], push_services)
                except Exception as e:
                    exception = e.__class__.__name__

                    JOB_FAILURES.labels(job_name, category, exception).inc()

                    aps.app.logger.warning("Job '" + job_name + "' failed. " + exception + ": " + str(e))
                    aps.app.logger.debug(traceback.format_exc())

                    if aps.app.config['SUSPEND_JOB_ON_FAILURE']:
                        aps.pause_job(job_name)

                        aps.app.logger.warning("Job '" + job_name + "' suspended.")

                    unset_job_collector_metrics(job_name)

    @staticmethod
    def read_prometheus_metrics(name):
        registry = get_registry(name)

        return generate_latest(registry)



import json
import logging
import traceback
import warnings

from datetime import datetime, timedelta

from apscheduler.events import EVENT_JOB_EXECUTED, EVENT_JOB_MAX_INSTANCES, EVENT_JOB_MISSED, EVENT_JOB_SUBMITTED
from apscheduler.schedulers.base import MaxInstancesReachedError

from flask import Flask

from flask_apialchemy import APIAlchemy

from flask_apscheduler import APScheduler

from flask_sqlalchemy import _EngineConnector as _EngineConnectorBase, SQLAlchemy as SQLAlchemyBase

from prometheus_client import generate_latest, Counter, Histogram
from prometheus_client.core import GaugeMetricFamily
from prometheus_client.utils import INF

from prometheus_flask_exporter import PrometheusMetrics

from pytz_deprecation_shim import PytzUsageWarning

from sqlalchemy.dialects import registry

from werkzeug.middleware.proxy_fix import ProxyFix

from .api.graphite import GraphiteBridge
from .api.pushgateway import Pushgateway
from .api.wavefront import Wavefront

from .config import config_by_name

from .util.prometheus_helper import prometheus_multiproc_dir, get_registry, register_collector, unregister_collector


class _EngineConnector(_EngineConnectorBase):
    def get_options(self, sa_url, echo):
        sa_url, options = super(_EngineConnector, self).get_options(sa_url, echo)

        if sa_url.drivername.startswith('oracle'):
            warnings.filterwarnings("ignore", ".*max_identifier_length.*")

            # options.setdefault('max_identifier_length', 30)

            options.setdefault('connect_args', {
                'events': True
            })

        poolclass = options.get('poolclass')

        if poolclass and poolclass.__name__ in ['NullPool', 'StaticPool']:
            if 'pool_size' in options.keys():
                options.pop('pool_size')

            if 'pool_recycle' in options.keys():
                options.pop('pool_recycle')

        return sa_url, options


class MisfireGraceTimeReachedError(Exception):
    def __init__(self, job):
        super(MisfireGraceTimeReachedError, self).__init__(
            'Job "%s" has reached its maximum grace time of (%d) seconds' %
            (job.id, job.misfire_grace_time))


class TimeoutWarning(RuntimeWarning):
    pass


class SQLAlchemy(SQLAlchemyBase):
    def make_connector(self, app=None, bind=None):
        """Creates the connector for a given state and bind."""
        return _EngineConnector(self, self.get_app(app), bind)


db = SQLAlchemy()
aa = APIAlchemy()

custom_dialect_info = {
    'bigquery': {
        'driver': 'pybigquery',
        'objname': 'BigQueryDialect'
    },
    'db2': {
        'driver': 'ibm_db',
        'objname': 'DB2Dialect'
    },
    'postgresql': {
        'driver': 'pg8000',
        'objname': 'PostgreSQLDialect'
    }
}

for dialect, info in custom_dialect_info.items():
    modulepath = '%s.database.%s.dialect' % (__package__, dialect)

    registry.register(dialect, modulepath, info['objname'])
    registry.register('%s.%s' % (dialect, info['driver']), modulepath, info['objname'])

warnings.filterwarnings("ignore", category=PytzUsageWarning)

aps = APScheduler()

app_registry_name = 'application'

app_registry = get_registry(app_registry_name)

metrics = PrometheusMetrics.for_app_factory(registry=app_registry)

default_job_category = 'internal'

JOB_EXECUTION_TIME = Histogram(
    'metrex_job_execution_seconds',
    'Time spent executing metREx job',
    ['job', 'category'],
    registry=app_registry,
    buckets=(.1, .25, .5, .75, 1.0, 2.5, 5.0, 7.5, 10.0, 30.0, 60.0, 150.0, 300.0, INF)
)

JOB_FAILURES = Counter(
    'metrex_job_failures_total',
    'Failed metREx job executions',
    ['job', 'category', 'exception'],
    registry=app_registry
)

registered_collectors = {}

job_collector_metrics = {}

job_push_service_names = {}


def create_app(config_name):
    from .. import blueprint

    app = Flask(__name__)

    app.wsgi_app = ProxyFix(app.wsgi_app)

    try:
        config_obj = config_by_name[config_name]()

        app.config.from_object(config_obj)

        if not app.debug:
            logger = logging.getLogger('werkzeug')
            logger.setLevel(logging.ERROR)

        db.init_app(app)
        aa.init_app(app)

        with app.app_context():
            config_obj.add_jobs_from_source(aa)

        app.config.from_object(config_obj)

        aps.init_app(app)
        metrics.init_app(app)

        app.register_blueprint(blueprint)
    except Exception as e:
        app.logger.critical(e)
        exit(1)

    return app


def delete_pushgateway_job_metrics(job_name, services):
    job_collector_registry = get_registry(job_name)

    grouping_key = get_grouping_key(job_name, job_collector_registry)

    if grouping_key is not None:
        for service, service_name in services.items():
            try:
                with aps.app.app_context():
                    dal = Pushgateway(aa)
                    dal.init_aa(service_name)

                    dal.client.delete(job=job_name, grouping_key=grouping_key)

                aps.app.logger.info("Removed metrics for job '" + job_name + "' from Pushgateway service '" + service + "'.")
            except Exception as e:
                aps.app.logger.warning("Failed removing metrics for job '" + job_name + "' from Pushgateway service '" + service + "'.")

                raise e


def get_grouping_key(job_name, job_collector_registry):
    if job_name in registered_collectors.keys():
        if registered_collectors[job_name].instance is None:
            generate_latest(job_collector_registry)

        grouping_key = {
            'job': job_name
        }

        if registered_collectors[job_name].instance:
            grouping_key['instance'] = registered_collectors[job_name].instance
    else:
        grouping_key = None

    return grouping_key


def get_jobs():
    job_list = aps.app.config.get('SCHEDULER_JOBS')

    return [
        job['id'] for job in job_list
    ]


def get_push_services(job_name, service_names=None):
    push_services = {}

    if service_names is not None:
        job_push_service_names[job_name] = service_names

    if job_name in job_push_service_names.keys():
        for vendor, services in aps.app.config['PUSH_SERVICES'].items():
            for service, service_name in services.items():
                if service_name in job_push_service_names[job_name]:
                    if vendor not in push_services.keys():
                        push_services[vendor] = {}

                    push_services[vendor][service] = service_name

    return push_services


def init_job_metric_counters(job_name, category):
    exceptions = [
        MaxInstancesReachedError,
        MisfireGraceTimeReachedError,
        TimeoutWarning
    ]

    for e in exceptions:
        JOB_FAILURES.labels(job_name, category, e.__name__)


def job_completed_event_listener(event):
    aps.app.logger.info("Job '" + event.job_id + "' completed.")


def job_not_executed_event_listener(event):
    job_list = aps.app.config.get('SCHEDULER_JOBS')

    for job in job_list:
        job_name = job['id']
        category = job['args'][0]

        if job_name == event.job_id:
            JOB_EXECUTION_TIME.labels(job_name, category).observe(0.0)

            exception = ''

            if event.code == EVENT_JOB_MAX_INSTANCES:
                exception = MaxInstancesReachedError.__name__
            elif event.code == EVENT_JOB_MISSED:
                exception = MisfireGraceTimeReachedError.__name__

            JOB_FAILURES.labels(job_name, category, exception).inc()
            break


def job_started_event_listener(event):
    aps.app.logger.info("Job '" + event.job_id + "' started.")


def push_graphite_job_metrics(job_name, services):
    job_collector_registry = get_registry(job_name)

    if job_name in registered_collectors.keys():
        if registered_collectors[job_name].instance is None:
            generate_latest(job_collector_registry)

        for service, service_name in services.items():
            try:
                with aps.app.app_context():
                    dal = GraphiteBridge(aa)
                    dal.init_aa(service_name)

                    dal.client.push(registry=job_collector_registry)

                aps.app.logger.info("Sent metrics for job '" + job_name + "' to Graphite bridge service '" + service + "'.")
            except Exception as e:
                aps.app.logger.warning("Failed sending metrics for job '" + job_name + "' to Graphite bridge service '" + service + "'.")

                raise e


def push_pushgateway_job_metrics(job_name, services):
    job_collector_registry = get_registry(job_name)

    grouping_key = get_grouping_key(job_name, job_collector_registry)

    if grouping_key is not None:
        for service, service_name in services.items():
            try:
                with aps.app.app_context():
                    dal = Pushgateway(aa)
                    dal.init_aa(service_name)

                    dal.client.push(job=job_name, registry=job_collector_registry, grouping_key=grouping_key)

                aps.app.logger.info("Sent metrics for job '" + job_name + "' to Pushgateway service '" + service + "'.")
            except Exception as e:
                aps.app.logger.warning("Failed sending metrics for job '" + job_name + "' to Pushgateway service '" + service + "'.")

                raise e


def push_wavefront_job_metrics(job_name, services):
    if job_name in job_collector_metrics.keys():
        default_source = ('prod' if aps.app.env == 'production' else 'dev') + '.monitoring.metrex'

        source = aps.app.config['WAVEFRONT_SOURCE'] or default_source

        for service, service_name in services.items():
            try:
                with aps.app.app_context():
                    dal = Wavefront(aa)
                    dal.init_aa(service_name)

                    dal.client.send(source=source, metrics=job_collector_metrics[job_name])

                aps.app.logger.info("Sent metrics for job '" + job_name + "' to Wavefront service '" + service + "'.")
            except Exception as e:
                aps.app.logger.warning("Failed sending metrics for job '" + job_name + "' to Wavefront service '" + service + "'.")

                raise e


def reload_job(job, run_on_reload=False):
    job_name = job['id']

    try:
        aps.remove_job(job_name)
        aps.add_job(**job)

        aps.app.logger.info("Job '" + job_name + "' reloaded.")
    except Exception as e:
        aps.app.logger.warning("Failed reloading job '" + job_name + "'.")

        raise e

    if run_on_reload:
        aps.run_job(job_name)


def set_job_collector_metrics(job_name, collector_metrics, push_services):
    job_collector_metrics[job_name] = collector_metrics

    if job_name not in registered_collectors.keys():
        use_timestamp = len(push_services) == 0

        registered_collectors[job_name] = JobCollector(job_name, use_timestamp)

        register_collector(job_name, registered_collectors[job_name])


def shutdown_scheduler():
    aps.shutdown()


def start_scheduler(config_name=None):
    aps.start(paused=True)

    aps.add_listener(job_not_executed_event_listener, EVENT_JOB_MAX_INSTANCES | EVENT_JOB_MISSED)
    aps.add_listener(job_started_event_listener, EVENT_JOB_SUBMITTED)
    aps.add_listener(job_completed_event_listener, EVENT_JOB_EXECUTED)

    if config_name is not None:
        initial_run_time = datetime.now()

        job_list = aps.app.config.get('SCHEDULER_JOBS')

        for job in job_list:
            job_name = job['id']
            category = job['args'][0]

            initial_run_time += timedelta(seconds=aps.app.config.get('SCHEDULER_JOB_DELAY_SECONDS'))

            aps.modify_job(job_name, next_run_time=initial_run_time)

            init_job_metric_counters(job_name, category)

        source_refresh_interval = aps.app.config.get('JOBS_SOURCE_REFRESH_INTERVAL')

        if source_refresh_interval is not None:
            source_refresh_job_name = 'SOURCE_REFRESH'

            job_kwargs = {
                'trigger': 'interval',
                'args': [default_job_category, config_name, source_refresh_job_name],
                'seconds': int(source_refresh_interval) * 60
            }

            aps.add_job(source_refresh_job_name, __package__ + ':update_jobs', **job_kwargs)

            init_job_metric_counters(source_refresh_job_name, default_job_category)

        aps.resume()


def unset_job_collector_metrics(job_name):
    if job_name in registered_collectors.keys():
        unregister_collector(job_name, registered_collectors[job_name])

        del registered_collectors[job_name]


def update_jobs(category, config_name, source_refresh_job_name):
    try:
        with JOB_EXECUTION_TIME.labels(source_refresh_job_name, category).time():
            config_obj = config_by_name[config_name]()

            with aps.app.app_context():
                config_obj.add_jobs_from_source(aa)

            job_list = aps.app.config.get('SCHEDULER_JOBS')

            jobs = {}

            for i in range(len(job_list)):
                job_name = job_list[i]['id']

                if job_name not in config_obj.jobs.keys():
                    push_services = get_push_services(job_name)

                    if 'pushgateway' in push_services.keys():
                        delete_pushgateway_job_metrics(job_name, push_services['pushgateway'])

                    unset_job_collector_metrics(job_name)

                    aps.remove_job(job_name)

                    aps.app.logger.info("Job '" + job_name + "' removed.")
                else:
                    jobs[job_name] = job_list[i]

            for i in range(len(config_obj.SCHEDULER_JOBS)):
                job_name = config_obj.SCHEDULER_JOBS[i]['id']

                run_job = True

                if job_name in jobs.keys():
                    run_job = json.dumps(config_obj.SCHEDULER_JOBS[i]) != json.dumps(jobs[job_name])

                    if run_job:
                        push_services = get_push_services(job_name)

                        if 'pushgateway' in push_services.keys():
                            delete_pushgateway_job_metrics(job_name, push_services['pushgateway'])

                        unset_job_collector_metrics(job_name)

                        aps.modify_job(**config_obj.SCHEDULER_JOBS[i])

                        aps.app.logger.info("Job '" + job_name + "' updated.")
                else:
                    aps.add_job(**config_obj.SCHEDULER_JOBS[i])

                    aps.app.logger.info("Job '" + job_name + "' added.")

                if run_job:
                    aps.run_job(job_name)

            aps.app.config.from_object(config_obj)
    except Exception as e:
        exception = e.__class__.__name__

        JOB_FAILURES.labels(source_refresh_job_name, category, exception).inc()

        aps.app.logger.warning("Job '" + source_refresh_job_name + "' failed. " + exception + ": " + str(e))
        aps.app.logger.debug(traceback.format_exc())

        # if aps.app.config.get('SUSPEND_JOB_ON_FAILURE'):
        #     aps.pause_job(source_refresh_job_name)
        #
        # aps.app.logger.warning("Job '" + source_refresh_job_name + "' suspended.")


class JobCollector:
    instance = None

    def __init__(self, job_name, use_timestamp):
        self._job_name = job_name
        self._use_timestamp = use_timestamp

    def collect(self):
        for metric_name, info in job_collector_metrics[self._job_name].items():
            json_label_data = list(info.keys())[0]

            label_dict = json.loads(json_label_data)

            g = GaugeMetricFamily(metric_name.replace('.', '_'), '', labels=label_dict.keys())

            if self.instance is None:
                if 'instance' in label_dict.keys():
                    self.instance = label_dict.pop('instance')
                else:
                    self.instance = ''

            for json_label_data, metric in info.items():
                label_dict = json.loads(json_label_data)

                options = {}

                if self._use_timestamp:
                    options['timestamp'] = metric[1]

                g.add_metric(label_dict.values(), metric[0], **options)

            yield g


__all__ = [
    'JOB_EXECUTION_TIME',
    'JOB_FAILURES',
    'TimeoutWarning',
    'aa',
    'app_registry_name',
    'aps',
    'db',
    'default_job_category',
    'metrics',
    'prometheus_multiproc_dir',
    'create_app',
    'generate_latest',
    'get_jobs',
    'get_push_services',
    'get_registry',
    'push_graphite_job_metrics',
    'push_pushgateway_job_metrics',
    'push_wavefront_job_metrics',
    'set_job_collector_metrics',
    'shutdown_scheduler',
    'start_scheduler',
    'unset_job_collector_metrics'
]



import json
import re

from urllib import parse

import pytz

from cryptofy import encoding, decrypt

from .misc_helper import str_to_bool


def build_bind_dict(binds, key=''):
    """Builds dict of connections to assign to SQLALCHEMY_BINDS"""
    bind_dict = {}

    encrypted_credentials = [
        'credentials_info',
        'password'
    ]

    for name, credentials in binds.items():
        if 'encrypted' in credentials.keys():
            if str_to_bool(credentials['encrypted']):
                if key != '':
                    for i in encrypted_credentials:
                        if i in credentials.keys():
                            credentials[i] = decrypt(bytes(key, encoding=encoding), credentials[i]).decode(encoding)
                else:
                    raise ValueError('No secret key found.')

        bind_dict[name] = build_dsn(credentials)

    return bind_dict


def build_dsn(credentials):
    """Constructs SQLAlchemy DSN string from credentials."""
    dialect_driver = [credentials['dialect']]

    if 'driver' in credentials.keys():
        dialect_driver.append(credentials['driver'])

    base = []

    if credentials['dialect'] == 'bigquery':
        project_dataset = []

        if 'project' in credentials.keys():
            project_dataset.append(credentials['project'])
        else:
            project_dataset.append('')

        if 'dataset' in credentials.keys():
            project_dataset.append(credentials['dataset'])
        else:
            if project_dataset[0] == '':
                project_dataset.append('')

        project_dataset_params = ['/'.join(project_dataset)]

        params = {}

        valid_params = [
            'location',
            'dataset_id',
            'arraysize',
            'credentials_path',
            'credentials_info',
            'default_query_job_config'
        ]

        for name in valid_params:
            if name in credentials.keys():
                if isinstance(credentials[name], dict):
                    params[name] = json.dumps(credentials[name], separators=(',', ':'))
                else:
                    params[name] = credentials[name]

        if len(params):
            project_dataset_params.append(parse.urlencode(params))

        base.append('?'.join(project_dataset_params))
    elif credentials['dialect'] == 'sqlite':
        if 'path' in credentials.keys():
            base.append(credentials['path'])
    else:
        hostname_port = [credentials['hostname']]

        if 'port' in credentials.keys():
            hostname_port.append(str(credentials['port']))

        if 'driver' in credentials.keys() and credentials['dialect'] == 'informix' and credentials['driver'] == 'ifx_jdbc':
            hostname_port_database_params = [':'.join(hostname_port)]

            if 'name' in credentials.keys():
                database_params = [credentials['name']]

                params = {
                    'delimident': 'y'
                }

                if 'server' in credentials.keys():
                    params['INFORMIXSERVER'] = credentials['server']

                if 'username' in credentials.keys():
                    params['user'] = credentials['username']

                if 'password' in credentials.keys():
                    params['password'] = credentials['password']

                database_params.append(';'.join(key + '=' + parse.quote(val) for key, val in params.items()))

                hostname_port_database_params.append(':'.join(database_params))

            base.append('/'.join(hostname_port_database_params))
        else:
            username_password = []

            if 'username' in credentials.keys():
                username_password.append(parse.quote(credentials['username']))

            if 'password' in credentials.keys():
                username_password.append(parse.quote(credentials['password']))

            if len(username_password):
                base.append(':'.join(username_password))

            if credentials['dialect'] == 'oracle':
                import cx_Oracle

                dsn = (cx_Oracle.makedsn(
                    host=credentials['hostname'],
                    port=str(credentials['port']),
                    service_name=credentials['name']
                ))

                base.append(dsn)
            else:
                hostname_port_database = [':'.join(hostname_port)]

                if 'name' in credentials.keys():
                    hostname_port_database.append(credentials['name'])

                hostname_port_database_params = ['/'.join(hostname_port_database)]

                if credentials['dialect'] in ['mysql', 'postgresql']:
                    params = {}

                    valid_params = [
                        'ssl_ca',
                        'ssl_cert',
                        'ssl_key'
                    ]

                    for name in valid_params:
                        if name in credentials.keys():
                            params[name] = credentials[name]

                    if len(params):
                        hostname_port_database_params.append(parse.urlencode(params))

                base.append('?'.join(hostname_port_database_params))

    uri = [
        '+'.join(dialect_driver),
        '@'.join(base)
    ]

    return '://'.join(uri)


def parse_services_for_binds(prefix, services):
    binds = {}

    service_name_pattern = re.compile(r'^' + re.escape(prefix) + r'(?P<name>.+)$', re.X)

    for service in services:
        m = service_name_pattern.match(service.name)

        if m is not None:
            components = m.groupdict()

            name = components['name']

            if 'timezone' in service.credentials.keys():
                if service.credentials['timezone'] not in pytz.all_timezones:
                    ValueError("Invalid timezone '" + service.credentials['timezone'] + "' defined for service '" + service.name + "'.")

            binds[name] = service.credentials

    return binds



from flask import make_response


def format_response(data, content_type):
    response = make_response(data)

    response.headers['Content-Type'] = content_type
    response.headers['Cache-Control'] = 'no-cache, no-store, must-revalidate'
    response.headers['Pragma'] = 'no-cache'
    response.headers['Expires'] = '0'

    return response



from flask_restx import Namespace


class MetricsDto:
    api = Namespace('metrics', description='Prometheus metrics')


class SchedulerDto:
    api = Namespace('scheduler', description='Flask-APScheduler API')



import re

from ssl import SSLContext

from sqlalchemy.dialects.postgresql.pg8000 import PGDialect_pg8000 as BaseDialect


class PostgreSQLDialect(BaseDialect):
    def _get_server_version_info(self, connection):
        v = connection.exec_driver_sql("select pg_catalog.version()").scalar()

        m = re.match(
            r".*(?:PostgreSQL|EnterpriseDB|(?:CockroachDB(?:\s+CCL)?)) "
            r"v?(\d+)\.?(\d+)?(?:\.(\d+))?(?:\.\d+)?(?:devel|beta)?",
            v
        )

        if not m:
            raise AssertionError(
                "Could not determine version from string '%s'" % v
            )

        return tuple([int(x) for x in m.group(1, 2, 3) if x is not None])

    def create_connect_args(self, url):
        cargs, cparams = super(PostgreSQLDialect, self).create_connect_args(url)

        if 'ssl_cert' in cparams.keys():
            ssl_context = SSLContext()

            if 'ssl_ca' in cparams.keys():
                ssl_context.load_verify_locations(cafile=cparams.pop('ssl_ca'))

            ssl_context.load_cert_chain(
                cparams.pop('ssl_cert'),
                keyfile=cparams.pop('ssl_key') if 'ssl_key' in cparams.keys() else None
            )

            cparams['ssl_context'] = ssl_context

        return cargs, cparams



from ibm_db_sa.ibm_db import dialect


class DB2Dialect(dialect):
    supports_statement_cache = False

    def get_isolation_level(self, connection):
        return 'CS'

    def create_connect_args(self, url):
        connect_args = super(DB2Dialect, self).create_connect_args(url)

        x = list(connect_args)
        y = list(x[0])

        y[0] += 'AUTHENTICATION=SERVER'

        x[0] = tuple(y)
        connect_args = tuple(x)

        return connect_args



from .. import ApiAccessLayer


class AppD(ApiAccessLayer):
    application = None
    client = None

    def init_aa(self, bind):
        super(AppD, self).init_aa(bind)

        self.application = self.service.application
        self.client = self.service.client



import os
import re

from sqlalchemy import text


class DatabaseAccessLayer:
    _session = None

    def __init__(self, db):
        self._db = db

    def _get_default_isolation_level(self, backend_name):
        isolation_level_env_var = re.sub(r'\s+', '_', backend_name).upper() + r'_ISOLATION_LEVEL'

        return os.getenv(isolation_level_env_var)

    def execute(self, statement):
        result = self._session.execute(text(statement))

        return result

    def init_db(self, bind):
        engine = self._db.get_engine(bind=bind)

        options = {}

        backend_name = engine.url.get_backend_name()

        default_isolation_level = self._get_default_isolation_level(backend_name)

        if default_isolation_level is not None:
            options['isolation_level'] = default_isolation_level

        self._session = self._db.create_scoped_session({
            'autocommit': False,
            'autoflush': False,
            'bind': engine.execution_options(**options)
        })



import os
import re
import yaml

import pytz

package_components = __package__.split('.')

package_components.pop()

job_func_root = '.'.join(package_components)


def apply_job_templates(jobs, templates):
    for job_name, credentials in jobs.items():
        if 'use_template' in credentials.keys():
            template_name = credentials.pop('use_template')

            if template_name in templates.keys():
                template = templates[template_name].copy()

                template.update(credentials)

                jobs[job_name].update(template)
            else:
                raise ValueError("Job '" + job_name + "' references undefined template: " + template_name + ".")

    return jobs


def build_job(category, name, bind, seconds, *args):
    return {
        'func': job_func_root + '.service.metrics_service:Metrics.generate_metrics',
        'trigger': 'interval',
        'args': (category, name, bind) + args,
        'id': name,
        'seconds': seconds
    }


def build_job_list(jobs, apialchemy_info, sqlalchemy_info):
    """Builds list of scheduled jobs to assign to SCHEDULER_JOBS"""
    job_list = []

    db_default_local_tz = os.getenv('DB_DEFAULT_LOCAL_TZ')

    if db_default_local_tz not in pytz.all_timezones:
        db_default_local_tz = 'UTC'

    apialchemy_prefix, apialchemy_binds = apialchemy_info
    sqlalchemy_prefix, sqlalchemy_binds = sqlalchemy_info

    push_service_vendors = [
        'graphite',
        'pushgateway',
        'wavefront'
    ]

    valid_service_names = list(apialchemy_binds.keys()) + list(sqlalchemy_binds.keys())

    service_name_pattern = re.compile(r'^' + r'(?:(?:' + re.escape(apialchemy_prefix) + r')|(?:' + re.escape(sqlalchemy_prefix) + r'))(?P<name>.+)$', re.X)

    for job_name, credentials in jobs.items():
        for k, v in credentials.items():
            if isinstance(v, str):
                undefined_env_vars, credentials[k] = populate_env_vars(v)

                if undefined_env_vars:
                    raise ValueError("Job '" + job_name + "' references undefined environment variable(s): " + ", ".join(undefined_env_vars) + ".")

        if 'services' in credentials.keys() and 'interval_minutes' in credentials.keys():
            service_names = {
                'source': [],
                'push': []
            }

            for service in credentials['services']:
                service_name = get_service_name(job_name, service, service_name_pattern, valid_service_names)

                if service_name in apialchemy_binds.keys():
                    if 'vendor' in apialchemy_binds[service_name].keys():
                        if apialchemy_binds[service_name]['vendor'] in push_service_vendors:
                            raise ValueError("Service '" + service + "' defined for job '" + job_name + "' references push API vendor.")
                    else:
                        raise ValueError("Service '" + service + "' defined for job '" + job_name + "' is missing vendor credential.")
                elif service_name not in sqlalchemy_binds.keys():
                    raise ValueError("Service '" + service + "' defined for job '" + job_name + "' does not match any defined connections.")

                service_names['source'].append(service_name)

            if 'push_services' in credentials.keys():
                for service in credentials['push_services']:
                    service_name = get_service_name(job_name, service, service_name_pattern, valid_service_names)

                    if service_name in apialchemy_binds.keys():
                        if 'vendor' in apialchemy_binds[service_name].keys():
                            if apialchemy_binds[service_name]['vendor'] not in push_service_vendors:
                                raise ValueError("Push service '" + service + "' defined for job '" + job_name + "' references unsupported API vendor.")
                        else:
                            raise ValueError("Push service '" + service + "' defined for job '" + job_name + "' is missing vendor credential.")
                    else:
                        raise ValueError("Push service '" + service + "' defined for job '" + job_name + "' does not match any defined API connections.")

                    service_names['push'].append(service_name)

            if all(service_name in apialchemy_binds.keys() for service_name in service_names['source']):
                vendor = None

                for service_name in service_names['source']:
                    if vendor is None:
                        vendor = apialchemy_binds[service_name]['vendor']

                    if apialchemy_binds[service_name]['vendor'] != vendor:
                        raise ValueError("Services defined for job '" + job_name + "' must reference same API vendor.")

                    if vendor == 'appdynamics':
                        if 'application' not in credentials.keys() and 'application' not in apialchemy_binds[service_name].keys():
                            raise ValueError("Services defined for job '" + job_name + "' must contain 'application' credential if job does not.")
                    elif vendor == 'newrelic':
                        if 'account_id' not in credentials.keys() and 'account_id' not in apialchemy_binds[service_name].keys():
                            raise ValueError("Services defined for job '" + job_name + "' must contain 'account_id' credential if job does not.")

                job_args = [
                    vendor,
                    job_name,
                    service_names,
                    int(credentials['interval_minutes']) * 60
                ]

                service_job_args = None

                if vendor == 'appdynamics':
                    application = None

                    if 'application' in credentials.keys():
                        application = credentials['application']

                    job_args.append(application)

                    if 'metric_path' in credentials.keys():
                        service_job_args = [
                            credentials['metric_path'],
                            int(credentials['interval_minutes'])
                        ]
                elif vendor == 'extrahop':
                    if 'metric_params' in credentials.keys() and 'metric_name' in credentials.keys() and 'aggregation' in credentials.keys():
                        service_job_args = [
                            credentials['metric_params'],
                            credentials['metric_name'],
                            credentials['aggregation'],
                            int(credentials['interval_minutes'])
                        ]
                elif vendor == 'newrelic':
                    account_id = None

                    if 'account_id' in credentials.keys():
                        account_id = int(credentials['account_id'])

                    job_args.append(account_id)

                    if 'statement' in credentials.keys() and 'value_attrs' in credentials.keys():
                        if isinstance(credentials['value_attrs'], str):
                            # For backward-compatibility with older versions, which expected comma-delimited string
                            value_attrs = list(filter(None, re.split(r'\s*,\s*', credentials['value_attrs'])))
                        else:
                            value_attrs = list(credentials['value_attrs'])

                        if not value_attrs:
                            raise ValueError("No value attributes specified for job '" + job_name + "'.")

                        service_job_args = [
                            credentials['statement'],
                            value_attrs,
                            int(credentials['interval_minutes'])
                        ]

                if service_job_args is not None:
                    job_args += service_job_args

                    if 'static_labels' in credentials.keys():
                        if isinstance(credentials['static_labels'], str):
                            # For backward-compatibility with older versions, which expected comma-delimited string of key:value pairs
                            static_label_list = list(filter(None, re.split(r'\s*,\s*', credentials['static_labels'])))

                            static_labels = [
                                tuple(filter(None, label_str.split(':'))) for label_str in static_label_list
                            ]
                        else:
                            static_labels = [
                                (label, value) for label, value in credentials['static_labels'].items()
                            ]

                        job_args.append(static_labels)

                    job_list.append(build_job(*job_args))

                    continue
            elif all(service_name in sqlalchemy_binds.keys() for service_name in service_names['source']):
                if 'statement' in credentials.keys() and 'value_columns' in credentials.keys():
                    if isinstance(credentials['value_columns'], str):
                        # For backward-compatibility with older versions, which expected comma-delimited string
                        value_columns = list(filter(None, re.split(r'\s*,\s*', credentials['value_columns'])))
                    else:
                        value_columns = list(credentials['value_columns'])

                    if not value_columns:
                        raise ValueError("No value columns specified for job '" + job_name + "'.")

                    job_args = [
                        'database',
                        job_name,
                        service_names,
                        int(credentials['interval_minutes']) * 60,
                        credentials['statement'],
                        value_columns
                    ]

                    if 'static_labels' in credentials.keys():
                        if isinstance(credentials['static_labels'], str):
                            # For backward-compatibility with older versions, which expected comma-delimited string of key:value pairs
                            static_label_list = list(filter(None, re.split(r'\s*,\s*', credentials['static_labels'])))

                            static_labels = [
                                tuple(filter(None, label_str.split(':'))) for label_str in static_label_list
                            ]
                        else:
                            static_labels = [
                                (label, value) for label, value in credentials['static_labels'].items()
                            ]

                        job_args.append(static_labels)

                    if 'timestamp_column' in credentials.keys():
                        job_args.append(credentials['timestamp_column'])

                        timezones = {
                            service_name: sqlalchemy_binds[service_name]['timezone'] if 'timezone' in sqlalchemy_binds[service_name].keys() else db_default_local_tz for service_name in service_names['source']
                        }

                        job_args.append(timezones)

                    job_list.append(build_job(*job_args))

                    continue
            else:
                raise ValueError("Services defined for job '" + job_name + "' must reference same connection type.")

        raise ValueError("Missing required credential(s) for job '" + job_name + "'.")

    return job_list


def get_jobs_from_services(job_prefix, template_prefix, services):
    jobs = {}
    templates = {}

    job_name_pattern = re.compile(r'^' + re.escape(job_prefix) + r'(?P<name>.+)$', re.X)
    template_name_pattern = re.compile(r'^' + re.escape(template_prefix) + r'.+$', re.X)

    for service in services:
        m = job_name_pattern.match(service.name)

        if m is not None:
            components = m.groupdict()

            name = components['name']

            jobs[name] = service.credentials
        else:
            m = template_name_pattern.match(service.name)

            if m is not None:
                templates[service.name] = service.credentials

    return apply_job_templates(jobs, templates)


def get_jobs_from_source(aa, apialchemy_info):
    jobs = {}

    apialchemy_prefix, apialchemy_binds = apialchemy_info

    service_name_pattern = re.compile(r'^' + r'(?:' + re.escape(apialchemy_prefix) + r')(?P<name>.+)$', re.X)

    api_vendor_pattern = re.compile(r'^(?:(?P<vendor>\w+)(?:\+(?:http|https|proxy))?)(?=://)', re.X)

    jobs_source_service = os.getenv('JOBS_SOURCE_SERVICE')

    if jobs_source_service is not None:
        m = service_name_pattern.match(jobs_source_service)

        if m is not None:
            components = m.groupdict()

            service_name = components['name']

            if service_name in apialchemy_binds.keys():
                conn_str = apialchemy_binds[service_name]

                m = api_vendor_pattern.match(conn_str)

                if m is not None:
                    components = m.groupdict()

                    if components['vendor'] == 'github':
                        from ..api.github import GitHub

                        jobs_source_org = os.getenv('JOBS_SOURCE_ORG')
                        jobs_source_repo = os.getenv('JOBS_SOURCE_REPO')

                        if jobs_source_org is not None and jobs_source_repo is not None:
                            jobs_source_path = os.getenv('JOBS_SOURCE_PATH')

                            if jobs_source_path is not None:
                                jobs_source_branch = os.getenv('JOBS_SOURCE_BRANCH', 'master')

                                dal = GitHub(aa)
                                dal.init_aa(service_name)

                                contents = dal.service.get_file_contents(
                                    dal.client,
                                    jobs_source_org,
                                    jobs_source_repo,
                                    jobs_source_path,
                                    jobs_source_branch
                                )

                                jobs = get_services_from_yaml(yaml.safe_load(contents))

                                jobs_source_templates_path = os.getenv('JOBS_SOURCE_TEMPLATES_PATH')

                                if jobs_source_templates_path is not None:
                                    contents = dal.service.get_file_contents(
                                        dal.client,
                                        jobs_source_org,
                                        jobs_source_repo,
                                        jobs_source_templates_path,
                                        jobs_source_branch
                                    )

                                    templates = get_services_from_yaml(yaml.safe_load(contents))

                                    jobs = apply_job_templates(jobs, templates)
                        else:
                            raise ValueError("Missing environment variable(s) required for jobs source service '" + jobs_source_service + "'.")
            else:
                raise ValueError("Service '" + jobs_source_service + "' not found.")

    return jobs


def get_service_name(job_name, service, service_name_pattern, valid_service_names):
    m = service_name_pattern.match(service)

    if m is not None:
        components = m.groupdict()

        service_name = components['name']

        if service_name not in valid_service_names:
            raise ValueError("Service '" + service + "' not found for job '" + job_name + "'.")
    else:
        raise ValueError("Invalid service name: '" + service + "'.")

    return service_name


def get_services_from_yaml(yaml_data):
    services = {}

    if isinstance(yaml_data, dict):
        for service, credentials in yaml_data.items():
            if isinstance(credentials, dict):
                services[service] = credentials

    return services


def populate_env_vars(string):
    undefined_env_vars = []

    env_var_pattern = re.compile(r'%%(?P<var>\w+)%%', re.X)

    for m in env_var_pattern.finditer(string):
        components = m.groupdict()

        env_var = os.getenv(components['var'])

        if env_var is None:
            undefined_env_vars.append(components['var'])
        else:
            current_env_var_pattern = re.compile(r'%%' + re.escape(components['var']) + r'%%', re.X)

            string = current_env_var_pattern.sub(env_var, string, 1)

    return undefined_env_vars, string



from flask import request
from flask_restx import Resource

from prometheus_client import CONTENT_TYPE_LATEST

from ...main import *

from ..service.metrics_service import Metrics

from ..util.dto import MetricsDto
from ..util.flask_helper import format_response

api = MetricsDto.api


def get_parser():
    parser = api.parser()
    parser.add_argument('job_id',
                        type=str,
                        help='The registered job name',
                        location='path')

    return parser


_parser = get_parser()


@api.route('')
class ApplicationMetrics(Resource):
    @api.doc('export_application_metrics')
    @metrics.do_not_track()
    def get(self):
        """Export application metrics"""
        if prometheus_multiproc_dir is None and request.environ.get('wsgi.multiprocess', False):
            api.abort(501, "Running in multiprocess mode but 'prometheus_multiproc_dir' env var not set.")

        result = Metrics.read_prometheus_metrics(app_registry_name)

        return format_response(result, CONTENT_TYPE_LATEST)


@api.route('/<job_id>')
@api.expect(_parser)
class JobMetrics(Resource):
    @api.doc('export_job_metrics')
    def get(self, job_id):
        """Export job metrics"""
        if prometheus_multiproc_dir is None and request.environ.get('wsgi.multiprocess', False):
            api.abort(501, "Running in multiprocess mode but 'prometheus_multiproc_dir' env var not set.")

        if job_id in get_jobs():
            result = Metrics.read_prometheus_metrics(job_id)

            return format_response(result, CONTENT_TYPE_LATEST)
        else:
            api.abort(404)



import os

from prometheus_client.core import CollectorRegistry
from prometheus_client.multiprocess import MultiProcessCollector

collector_registries = {}

prometheus_multiproc_dir = os.getenv('PROMETHEUS_MULTIPROC_DIR')


def get_registry(name):
    if name not in collector_registries.keys():
        collector_registries[name] = CollectorRegistry()

        if prometheus_multiproc_dir is not None:
            path = os.path.join(prometheus_multiproc_dir, name.lower())

            if os.path.isdir(prometheus_multiproc_dir) and not os.path.isdir(path):
                os.mkdir(path)

            MultiProcessCollector(registry=collector_registries[name], path=path)

    return collector_registries[name]


def register_collector(name, collector):
    job_registry = get_registry(name)

    job_registry.register(collector)


def unregister_collector(name, collector):
    if name in collector_registries.keys():
        collector_registries[name].unregister(collector)

        del collector_registries[name]



import os
import re

from cryptofy import encoding, decrypt

from urllib import parse

from .misc_helper import str_to_bool


def build_conn_str(credentials):
    """Constructs API connection string from credentials."""
    uri = []

    if 'vendor' in credentials.keys():
        base = []

        if 'apikey' in credentials.keys():
            base.append(parse.quote(credentials['apikey']))
        else:
            if credentials['vendor'] == 'appdynamics':
                username_account = []
    
                if 'username' in credentials.keys():
                    username_account.append(parse.quote(credentials['username']))
    
                    if 'account' in credentials.keys():
                        username_account.append(parse.quote(credentials['account']))
    
                if len(username_account):
                    username_account_password = [
                        '@'.join(username_account)
                    ]
    
                    if 'password' in credentials.keys():
                        username_account_password.append(parse.quote(credentials['password']))
    
                    base.append(':'.join(username_account_password))
            else:
                username_password = []
    
                if 'username' in credentials.keys():
                    username_password.append(parse.quote(credentials['username']))
    
                    if 'password' in credentials.keys():
                        username_password.append(parse.quote(credentials['password']))
    
                if len(username_password):
                    base.append(':'.join(username_password))

        hostname_port = []

        scheme = [credentials['vendor']]

        if 'hostname' in credentials.keys():
            url_components = parse.urlparse(credentials['hostname'])

            if url_components.scheme:
                scheme.append(url_components.scheme)

            if url_components.hostname is not None:
                hostname = url_components.hostname
            else:
                hostname = credentials['hostname']

            hostname_port.append(hostname)

        if 'port' in credentials.keys():
            hostname_port.append(str(credentials['port']))

        hostname_port_path = []

        if len(hostname_port):
            hostname_port_path.append(':'.join(hostname_port))

            if credentials['vendor'] == 'appdynamics':
                if 'application' in credentials.keys():
                    hostname_port_path.append(parse.quote(credentials['application']))
            elif credentials['vendor'] == 'newrelic':
                if 'account_id' in credentials.keys():
                    hostname_port_path.append(str(credentials['account_id']))
            elif credentials['vendor'] in ['graphite', 'wavefront']:
                if 'prefix' in credentials.keys():
                    hostname_port_path.append(parse.quote(credentials['prefix']))

        if len(hostname_port_path):
            base.append('/'.join(hostname_port_path))

        uri.append('+'.join(scheme))

        if len(base):
            uri.append('@'.join(base))

    return '://'.join(uri)


def build_bind_dict(binds, key=''):
    """Builds dict of connections to assign to APIALCHEMY_BINDS"""
    bind_dict = {}

    encrypted_credentials = [
        'apikey',
        'password'
    ]

    for name, credentials in binds.items():
        if 'encrypted' in credentials.keys():
            if str_to_bool(credentials['encrypted']):
                if key != '':
                    for i in encrypted_credentials:
                        if i in credentials.keys():
                            credentials[i] = decrypt(bytes(key, encoding=encoding), credentials[i]).decode(encoding)
                else:
                    raise ValueError('No secret key found.')

        bind_dict[name] = build_conn_str(credentials)

    return bind_dict


def get_default_push_service_names(push_services):
    default_push_service_names = []

    default_push_services = list(filter(None, re.split(r'\s*,\s*', os.getenv('SEND_ALL_JOBS_TO_SERVICES', ''))))

    for services in push_services.values():
        for service, service_name in services.items():
            if service in default_push_services and service_name not in default_push_service_names:
                default_push_service_names.append(service_name)

    return default_push_service_names


def get_push_services(prefix, services):
    push_services = {}

    service_name_pattern = re.compile(r'^' + re.escape(prefix) + r'(?P<name>.+)$', re.X)

    vendors = {
        'graphite': 'Graphite bridge',
        'pushgateway': 'Pushgateway',
        'wavefront': 'Wavefront'
    }

    for service in services:
        m = service_name_pattern.match(service.name)

        if m is not None:
            components = m.groupdict()

            name = components['name']

            if 'vendor' in service.credentials.keys():
                vendor = service.credentials['vendor']

                if vendor in vendors.keys():
                    if vendor not in push_services.keys():
                        push_services[vendor] = {}

                    push_services[vendor][service.name] = name

    return push_services


def parse_services_for_binds(prefix, services):
    binds = {}

    service_name_pattern = re.compile(r'^' + re.escape(prefix) + r'(?P<name>.+)$', re.X)

    for service in services:
        m = service_name_pattern.match(service.name)

        if m is not None:
            components = m.groupdict()

            name = components['name']

            binds[name] = service.credentials

    return binds



import os

from flask import send_from_directory
from flask_script import Manager

if __package__ is None:
    import sys

    __package__ = os.path.basename(sys.path[0])

    sys.path.append(os.path.dirname(sys.path[0]))

from .app.main import *


config_name = os.getenv('BOILERPLATE_ENV', 'dev')

app = create_app(config_name)

manager = Manager(app)

host = os.getenv('CF_INSTANCE_INTERNAL_IP', os.getenv('IP_ADDRESS', '127.0.0.1'))
port = int(os.getenv('PORT', 5000))


@app.route('/favicon.ico')
def favicon():
    return send_from_directory(
        os.path.join(app.root_path, 'static'),
        'favicon.ico',
        mimetype='image/vnd.microsoft.icon'
    )


@manager.command
def run():
    if not (app.debug and os.getenv('WERKZEUG_RUN_MAIN') is None):
        start_scheduler(config_name)

    app.run(host=host, port=port)


if __name__ == '__main__':
    manager.run()



from .. import ApiAccessLayer


class NewRelic(ApiAccessLayer):
    account_id = None
    client = None

    def init_aa(self, bind):
        super(NewRelic, self).init_aa(bind)

        self.account_id = self.service.account_id
        self.client = self.service.client



import re

from cfenv import AppEnv

from flask import Blueprint, url_for
from flask_restx import Api

from .main.controller.metrics_controller import api as metrics_ns
from .main.controller.scheduler_controller import api as scheduler_ns

__title__ = 'metREx'
__version__ = '0.7.2'
__description__ = 'SQL query and monitoring system metrics exporter for Prometheus'

blueprint = Blueprint('api', __name__)
blueprint.config = {}


@blueprint.record
def record_config(setup_state):
    app = setup_state.app
    blueprint.config = dict([(key, value) for (key, value) in app.config.items()])


class MyApi(Api):
    '''
    Extension of the main entry point for the application.
    Need to modify Swagger API behavior to support running in HTTPS when deployed to PCF
    '''
    @property
    def specs_url(self):
        '''
        The Swagger specifications absolute url (ie. `swagger.json`)

        :rtype: str
        '''
        return url_for(self.endpoint('specs'), _external=False)


env = AppEnv()

kwargs_api = {
    'title': __title__,
    'version': __version__,
    'description': __description__
}

if env.name is not None:
    pattern = re.compile(
        r"""
            (?P<title>.*)(?=-v[0-9.+]+$)
            (?:
                -v(?P<version>[0-9.+]+)$
            )
        """,
        re.X
    )

    m = pattern.match(env.name)

    if m is not None:
        components = m.groupdict()

        kwargs_api.update(components)
    else:
        kwargs_api['title'] = env.name

api = MyApi(blueprint, **kwargs_api)

api.add_namespace(metrics_ns, path='/metrics')
api.add_namespace(scheduler_ns, path='/scheduler')


@api.errorhandler
def default_error_handler(error):
    '''
    Default error handler
    '''
    message = str(error)

    if 'ERROR_INCLUDE_MESSAGE' in api.app.config.keys():
        if not api.app.config['ERROR_INCLUDE_MESSAGE']:
            message = 'Internal Server Error'

    return {'message': message}, 500



from .. import ApiAccessLayer


class Wavefront(ApiAccessLayer):
    client = None

    def init_aa(self, bind):
        super(Wavefront, self).init_aa(bind)

        self.client = self.service.client



import json
import os
import yaml

from apscheduler.executors.pool import ProcessPoolExecutor, ThreadPoolExecutor

from cfenv import AppEnv

from dotenv import load_dotenv

from sqlalchemy.pool import NullPool, SingletonThreadPool

from .util import api_helper
from .util import apscheduler_helper
from .util.misc_helper import str_to_bool
from .util import sqlalchemy_helper


def build_vcap_services_from_dict(data):
    vcap_services = {
        'user_provided': []
    }

    if isinstance(data, dict):
        for service, credentials in data.items():
            if isinstance(credentials, dict):
                vcap_services['user_provided'].append({
                    'name': service,
                    'label': 'user_provided',
                    'credentials': credentials
                })

    return json.dumps(vcap_services, separators=(', ', ': '), sort_keys=True)


def get_real_file_path(file_path):
    return os.path.join(basedir, file_path)


def get_secret_key_from_file(file_path, default_file_path):
    secret_key = ''

    if file_path is None:
        if os.path.isfile(get_real_file_path(default_file_path)):
            file_path = default_file_path

    if file_path is not None:
        yaml_data = read_yaml_from_file(file_path)

        if 'secret-key' in yaml_data.keys():
            secret_key = yaml_data['secret-key']

    return secret_key


def get_services_from_file(file_path, default_file_path):
    services = {}

    if file_path is None:
        if os.path.isfile(get_real_file_path(default_file_path)):
            file_path = default_file_path

    if file_path is not None:
        yaml_data = read_yaml_from_file(file_path)

        services = build_vcap_services_from_dict(yaml_data)

    return services


def read_yaml_from_file(file_path):
    try:
        with open(get_real_file_path(file_path), 'r') as stream:
            data = yaml.safe_load(stream)
    except Exception as err:
        raise ValueError(err)

    return data


available_cpu_count = len(os.sched_getaffinity(0))

basedir = os.path.abspath(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))

env_path = os.path.join(basedir, '.env')

if os.path.isfile(env_path):
    load_dotenv(env_path)

app_prefix = 'METREX'

service_prefix = {
    'APIALCHEMY': os.getenv('API_PREFIX', app_prefix + '_API_'),
    'SQLALCHEMY': os.getenv('DB_PREFIX', app_prefix + '_DB_')
}

job_prefix = os.getenv('JOB_PREFIX', app_prefix + '_JOB_')
template_prefix = os.getenv('JOB_PREFIX', app_prefix + '_TEMPLATE_')


class Config:
    DEBUG = str_to_bool(os.getenv('DEBUG', True))
    TESTING = False
    SECRET_KEY = os.getenv('SECRET_KEY', get_secret_key_from_file(os.getenv('SECRET_PATH'), 'secrets.yml'))

    ERROR_INCLUDE_MESSAGE = str_to_bool(os.getenv('ERROR_INCLUDE_MESSAGE', True))
    RESTPLUS_MASK_SWAGGER = False
    SWAGGER_UI_REQUEST_DURATION = True

    APIALCHEMY_BINDS = {}

    SQLALCHEMY_BINDS = {}

    SQLALCHEMY_ENGINE_OPTIONS = {
        'poolclass': NullPool,
        'pool_pre_ping': True,
        'pool_reset_on_return': None
    }

    SQLALCHEMY_TRACK_MODIFICATIONS = False

    PUSH_SERVICES = {}

    DEFAULT_PUSH_SERVICE_NAMES = []

    WAVEFRONT_SOURCE = os.getenv('WAVEFRONT_SOURCE')

    JOBS_SOURCE_REFRESH_INTERVAL = None

    SCHEDULER_JOBS = []

    SCHEDULER_API_ENABLED = False

    SCHEDULER_EXECUTORS = {
        'default': ThreadPoolExecutor(max_workers=int(os.getenv('MAX_WORKERS', '10'))),
        'processpool': ProcessPoolExecutor(max_workers=int(os.getenv('MAX_WORKERS', str(available_cpu_count))))
    }

    SCHEDULER_JOB_DEFAULTS = {
        'executor': 'processpool' if os.getenv('JOB_EXECUTOR', 'ThreadPool').lower() == 'processpool' else 'default',
        'coalesce': True,
        'max_instances': int(os.getenv('MAX_INSTANCES', '1')),
        'misfire_grace_time': int(os.getenv('MISFIRE_GRACE_TIME', '15')),
        'replace_existing': True
    }

    SCHEDULER_JOB_DELAY_SECONDS = int(os.getenv('JOB_INITIAL_DELAY_SECONDS', '15'))

    LIMIT_JOB_EXECUTION_TIME = str_to_bool(os.getenv('LIMIT_JOB_EXECUTION_TIME', False))

    SUSPEND_JOB_ON_FAILURE = str_to_bool(os.getenv('SUSPEND_JOB_ON_FAILURE', False))

    def __init__(self):
        if os.getenv('VCAP_SERVICES') is None:
            os.environ['VCAP_SERVICES'] = get_services_from_file(os.getenv('SERVICES_PATH'), 'env/services.yml')

        env = AppEnv()

        self.apialchemy_binds = api_helper.parse_services_for_binds(service_prefix['APIALCHEMY'], env.services)

        self.sqlalchemy_binds = sqlalchemy_helper.parse_services_for_binds(service_prefix['SQLALCHEMY'], env.services)

        self.push_services = api_helper.get_push_services(service_prefix['APIALCHEMY'], env.services)

        self.default_push_service_names = api_helper.get_default_push_service_names(self.push_services)

        if os.getenv('JOBS_SOURCE_SERVICE') is not None:
            self.JOBS_SOURCE_REFRESH_INTERVAL = os.getenv('JOBS_SOURCE_REFRESH_INTERVAL', '60')

        self.service_jobs = apscheduler_helper.get_jobs_from_services(job_prefix, template_prefix, env.services)
        self.source_jobs = {}
        self.jobs = self.service_jobs

    def add_jobs_from_source(self, aa):
        if self.JOBS_SOURCE_REFRESH_INTERVAL is not None:
            self.source_jobs = apscheduler_helper.get_jobs_from_source(aa, (service_prefix['APIALCHEMY'], self.APIALCHEMY_BINDS))
            self.jobs = {**self.service_jobs, **self.source_jobs}

    @property
    def apialchemy_binds(self):
        return self._apialchemy_binds

    @apialchemy_binds.setter
    def apialchemy_binds(self, value):
        self._apialchemy_binds = value

        self.APIALCHEMY_BINDS = api_helper.build_bind_dict(
            self._apialchemy_binds,
            self.SECRET_KEY
        )

    @property
    def sqlalchemy_binds(self):
        return self._sqlalchemy_binds

    @sqlalchemy_binds.setter
    def sqlalchemy_binds(self, value):
        self._sqlalchemy_binds = value

        self.SQLALCHEMY_BINDS = sqlalchemy_helper.build_bind_dict(
            self._sqlalchemy_binds,
            self.SECRET_KEY
        )

    @property
    def push_services(self):
        return self._push_services

    @push_services.setter
    def push_services(self, value):
        self._push_services = value

        self.PUSH_SERVICES = self._push_services

    @property
    def default_push_service_names(self):
        return self._default_push_service_names

    @default_push_service_names.setter
    def default_push_service_names(self, value):
        self._default_push_service_names = value

        self.DEFAULT_PUSH_SERVICE_NAMES = self._default_push_service_names

    @property
    def jobs(self):
        return self._jobs

    @jobs.setter
    def jobs(self, value):
        self._jobs = value

        self.SCHEDULER_JOBS = apscheduler_helper.build_job_list(
            self._jobs,
            (service_prefix['APIALCHEMY'], self._apialchemy_binds),
            (service_prefix['SQLALCHEMY'], self._sqlalchemy_binds)
        )


class DevelopmentConfig(Config):
    ENV = 'development'


class TestingConfig(Config):
    ENV = 'development'
    TESTING = True
    PRESERVE_CONTEXT_ON_EXCEPTION = False

    SQLALCHEMY_ENGINE_OPTIONS = {
        'poolclass': SingletonThreadPool
    }

    LIMIT_JOB_EXECUTION_TIME = False

    def __init__(self):
        database_service_name = service_prefix['SQLALCHEMY'] + 'TEST'

        services = {
            database_service_name: {
                'dialect': 'sqlite'
            },
            job_prefix + 'QUERY': {
                'services': [
                    database_service_name
                ],
                'interval_minutes': 0,
                'statement': 'SELECT value AS "metric", label1, label2, ts FROM metrics',
                'value_columns': [
                    'metric'
                ],
                'static_labels': {
                    'static': 'test'
                },
                'timestamp_column': 'ts'
            }
        }

        os.environ['VCAP_SERVICES'] = build_vcap_services_from_dict(services)

        if os.getenv('JOBS_SOURCE_SERVICE') is not None:
            del os.environ['JOBS_SOURCE_SERVICE']

        super(TestingConfig, self).__init__()


class ProductionConfig(Config):
    DEBUG = str_to_bool(os.getenv('DEBUG', False))

    ERROR_INCLUDE_MESSAGE = str_to_bool(os.getenv('ERROR_INCLUDE_MESSAGE', False))


config_by_name = dict(
    dev=DevelopmentConfig,
    test=TestingConfig,
    prod=ProductionConfig
)
